'''
Tower Tennis
    by Andrew Eshelman, Amanda White, and Sajjan Singh Mehta
    Recitation Assignment #8
    12/06/2007
    
A version of breakout in which the player moves up a tower, moving
and breaks through series of obstructions in the form of bricks.  
'''

from visual import *
from random import random
from pick import *
from score import *

# Runtime constants
dt = 0.005              # Timestep
theta = 0               # Angle of the paddle
dtheta = pi / 50        # Constant change in angle from user input
r = 2                   # Radius of paddle and used for distance from paddle

# Runtime variables
starting = 1            # Shows the starting screen
running = 1             # Tells whether the game is running
stall = 0               # Skips interactions to prevent errors
pause = 0               # Indicates whether the game has been paused
score = 0               # Score based on blocks hit and height
counter = 0             # Keeps track of the number of patterns generated
speed = 1               # Speed of the game run
patterns = []           # Patterns of block obstacles stored
obstacles = []          # List of the block obstacles 
starttime = time.time() # System time of game start

# Physics values
g = vector(0, -10)      # Gravitational acceleration
m = 5                   # Mass of the game ball
bounce_v = 2 * mag(g)   # Velocity from force applied by the paddle if the ball is too slow
inelasticity = 0.95     # Coefficient of restitution for inelastic collisions

# Grab name for highscore usage
print 'Please enter your name (used for global high scores):'
name = raw_input()
name = string.replace(name, ' ', '')

# Scene setup
scene = display(title = 'Tower Tennis', width = 336, height = 600,
                range = (14, 25, 10), fov = dtheta, scale = (0.04, 0.04, 0.1))

# Game ball
ball = sphere(radius = 0.5, pos = vector(0, 0), v = vector(0, 0))

# Game Paddle
paddle = frame(pos = vector(0, -10), axis = (0, 1, -4))
target = cylinder(frame = paddle, pos = (0, 0), axis = (0, 0.1), radius = r, color = color.red)
ring(frame = paddle, pos = (0, -0.1), axis = (0, 0.1), radius = r, color = color.blue, thickness = 0.05)
ellipsoid(frame = paddle, pos = (-r, -0.5), size = (4 * r, 1, 1), color = (0.55, 0.27, 0.07))

# Arrow seen when ball goes out of visible frame
arrow = arrow(pos = (0, 23), axis = (0, 2))
arrow.visible = 0

# High scores
highscore_name, highscore = getHighScores()

# Counter and score labels
score_label = label(pos = (10, -23), linecolor = color.black, height = 11,
                    color = color.white, text = 'Score: '+ str(score))
time_label = label(pos = (-10, -23), linecolor = color.black, height = 11,
                   color = color.white, text = 'Time: 0')
pause_label = label(pos = (0, 0), linecolor = color.black, height = 20,
                    color = color.green, text = 'P A U S E D', visible = 0)
highscore_label = label(pos = (0, -21), linecolor = color.black, height = 11,
                   color = color.white, 
                   text = 'Highscore: '+ str(highscore) +' - '+ highscore_name)

# Scene properties
scene.autoscale = 0
scene.userspin = 0
scene.userzoom = 0

# Processes user input from the mouse and keyboard
def checkUserInput():
    global scene, theta, dtheta, paddle, pause

    if scene.mouse.events:
        e = scene.mouse.getevent()
        if e.button == 'left' and theta < (pi / 12):
            theta += dtheta
            paddle.rotate(angle = dtheta, axis = vector(0, 0, 1))
        elif e.button == 'right' and theta > (-pi / 12):
            theta -= dtheta
            paddle.rotate(angle = -dtheta, axis = vector(0, 0, 1))
    
    if scene.kb.keys:
        key = scene.kb.getkey()
        
        if (key == 'a' or key == 'left') and theta < (pi / 12):
            theta += dtheta
            paddle.rotate(angle = dtheta, axis = vector(0, 0, 1))
        if (key == 'd' or key == 'right') and theta > (-pi / 12):
            theta -= dtheta
            paddle.rotate(angle = -dtheta, axis = vector(0, 0, 1))
        if key == 'p':
            pause = not pause

def restartGame():
    global starting, running, pause, score, counter, speed, g, starttime, ball, paddle, patterns, obstacles, nextframe, nextblocks, theta, score_label

    # Reset game values
    starting = 1
    running = 1
    stall = 0
    pause = 0
    score = 0
    counter = 0
    speed = 1
    g = vector(0, -10)
    starttime = time.time()
    score_label.text = 'Score: '+ str(score)
    
    ball.pos = vector(0, 0)
    ball.v = vector(0, 0)
    paddle.pos = vector(0, -10)
    paddle.rotate(angle = -theta, axis = vector(0, 0, 1))
    theta = 0
    
    # Horrible deletion of blocks - way too messy
    for i in xrange(len(patterns)):
        for j in patterns[i].objects:
            j.visible = 0
    for i in xrange(len(nextframe)):
        for j in nextframe[i].objects:
            j.visible = 0
    for i in xrange(len(obstacles)):
        for j in xrange(len(obstacles[0])):
            obstacles[0][0].block.visible = 0
            del obstacles[0][0]
        del obstacles[0]
    for i in xrange(len(nextblocks)):
        for j in xrange(len(nextblocks[0])):
            del nextblocks[0][0]
        del nextblocks[0]
    
    for i in xrange(len(patterns)):
        del patterns[0]
    for i in xrange(len(nextframe)):
        del nextframe[0]
    

# Handle loss of game
def promptUserAction():
    global name, score, highscore, highscore_name, highscore_label
    
    # Update High Scores
    if score > 10:
        postHighScore(name, score)
    highscore, highscore_name = getHighScores()
    highscore_label.text = 'Highscore: '+ str(highscore) + ' - '+ highscore_name
    
    if score > highscore:
        loss_label = label(pos = (0, 5), linecolor = color.black, height = 16,
                           color = color.green, 
                           text = 'Well Done! You got the highscore!\nWould you like to play again?')
    else:
        loss_label = label(pos = (0, 5), linecolor = color.black, height = 16,
                           color = color.green, 
                           text = 'Sorry, you lost.\nWould you like to play again?')

    cont = label(pos = (-5, 0), height = 11, linecolor = color.blue,
                 color = color.white, font = 'Arial', text = 'Continue')
    cont_box = box(pos = vector(-5, 0), length = 4, height = 2, width = 0.1,
                   color = color.black)
    quit = label(pos = (5, 0), height = 11, linecolor = color.blue,
                 color = color.white, font = 'Arial', text = 'Quit')
    quit_box = box(pos = vector(5, 0), length = 3, height = 2, width = 0.1,
               color = color.black)
    open = label(pos = (0, -3), height = 11, linecolor = color.blue,
                 color = color.white, font = 'Arial', text = 'Open HighScores')
    open_box = box(pos = vector(0, -3), length = 8, height = 2, width = 0.1,
               color = color.black)
               
    while 1:
        if scene.mouse.clicked:
            e = scene.mouse.getclick()
            if e.pick == cont_box:
                cont.visible = cont_box.visible = 0
                quit.visible = quit_box.visible = 0
                open.visible = open_box.visible = 0
                loss_label.visible = 0
                return 1
            elif e.pick == quit_box:
                cont.visible = cont_box.visible = 0
                quit.visible = quit_box.visible = 0
                open.visible = open_box.visible = 0
                loss_label.visible = 0
                return 0
            elif e.pick == open_box:
                openHighScores()


while 1:
    global highscore, highscore_name
    
    # Handle incoming events
    checkUserInput()
    
    if starting:
        highscore, highscore_name = getHighScores()
    
        # Wait for user to start the game
        start = label(pos = (0, -8), color = color.green, linecolor = color.black,
                      text = 'Point your mouse over the paddle to start!')
        while 1:
            if scene.mouse.pick == target:
                start.visible = 0
                starting = 0
                break
    
    # Handle not running or paused conditions
    if running == 0:
        if promptUserAction():
            restartGame()
            continue
        else:
            print 'Final Score: '+ str(score) +'.  Thanks for playing!'
            scene.visible = 0
            sys.exit(0)
    elif pause:
        pause_label.visible = 1
        continue
    else:
        pause_label.visible = 0
    
    time_label.text = 'Time: '+ str(int(time.time() - starttime))

    rate(1.5 / dt)    # Set the animation rate
    stall += 1        # Update the stall
    paddle.pos = scene.mouse.pos
    
    # Calculate the ball's new position and velocity
    ball.v += g * dt
    ball.pos += ball.v * dt
    
    # Update the pattern positions
    for p in patterns:
        p.pos.y -= 0.005
        
        if p.y + p.height <= -25:
            i = patterns.index(p)
            
            for j in obstacles[i]:
                j.block.visible = 0
            del obstacles[i]
            del patterns[i]
    
    # Add another pattern to the game if necessary
    if len(patterns) == 0 or (len(patterns) == 1 and patterns[0].pos.y + patterns[0].height <= 0):
        if len(patterns) == 0:
            pattern, blocks = pick('initial')
        elif counter % 5 == 0:
            pattern, blocks = pick('special')
        else:
            pattern, blocks = pick()
         
        #pattern.pos = (0, 25)# + pattern.height)
        patterns.append(pattern)
        obstacles.append(blocks)
        
        counter += 1
    
    
    # If the wait period has expired, check for any ball interactions 
    if stall > 5:
        # Calculate transformed ball and paddle positions
        ball_rot = rotate(ball.pos, angle = -theta)
        paddle_rot = rotate(paddle.pos, angle = -theta)
        
        # Display an arrow when the ball goes off of the screen
        if ball.pos.y > 25:
            arrow.visible = 1
            arrow.pos.x = ball.pos.x
        else:
            arrow.visible = 0
        
        # Handle paddle hit
        if abs(ball_rot.y - paddle_rot.y) < 0.1 and abs(ball_rot.x - paddle_rot.x) <= r:
            # Rotate the velocity vector and reverse the y diretion,
            # considering an inelastic collision
            ball.v = rotate(ball.v, angle = -theta)
            ball.v.y *= -1 * inelasticity
            if abs(theta) > 0:
                ball.v.x *= abs(sin(theta)) * inelasticity
            ball.v = rotate(ball.v, angle = theta)
            
            # Add a bounce force to the ball if below a certain velocity
            if abs(ball.v.y) < bounce_v:
                ball.v.y = bounce_v * ball.v.y / abs(ball.v.y)
            stall = 0
        
        # Handle hits against the walls
        if abs(ball.pos.x) > 14:
            ball.v.x *= -1 * inelasticity
            stall = 0
        ''' Allow bouncing on the bottom of the screen
        if ball.pos.y < -25:
            ball.v.y *= -1 * inelasticity
            stall = 0
        '''
        # End the game if the user hits the bottom
        if ball.pos.y <= -25:
            pause = 1
            running = 0
        
        # Check for interactions of the ball with obstacles
        for i in obstacles:
            block_num = range(len(i))
            for j in block_num:
                ob = i[j]
                
                # Properties of the individual brick obstacle
                pos = ob.pos + ob.frame.pos
                h = ob.height / 2.0
                w = ob.length / 2.0
                
                # If block is below the bottom of the window, delete it
                if pos.y < -25:
                    ob.block.visible = 0
                    del i[j]
                    del block_num[len(block_num) - 1]
                
                # If block is above the top of the window, do not interact
                if pos.y > 25:
                    continue
                
                if mag(ball.pos - pos) < sqrt(h**2 + w**2 + r**2):
                    if ball.pos.y > pos.y - h and ball.pos.x > pos.x - w and ball.pos.x < pos.x + w and ball.pos.y < pos.y + h:
                        if w - abs(ball.pos.x - pos.x) > h - abs(ball.pos.y - pos.y):
                            ball.v.y *= -.95
                            score += ob.handleHit()
                            if ob.getStrength() == 0:
                                del i[j]
                            
                            score_label.text = 'Score: '+ str(score)
                            stall = 0
                            break
                        else:
                            ball.v.x *= -.95
                            score += ob.handleHit()
                            if ob.getStrength() == 0:
                                del i[j]
                            
                            score_label.text = 'Score: '+ str(score)
                            stall = 0
                            break
